import os

import psycopg_pool
from fastapi import FastAPI, Response, HTTPException
from fastapi.responses import RedirectResponse
from fastapi.middleware.cors import CORSMiddleware

import tilekiln
from tilekiln.kiln import Kiln
from tilekiln.config import Config
from tilekiln.tile import Tile


# Constants for environment variable names
TILEKILN_CONFIG = "TILEKILN_CONFIG"
TILEKILN_URL = "TILEKILN_URL"
TILEKILN_ID = "TILEKILN_ID"

STANDARD_HEADERS = {"Cache-Control": "no-cache"}

kiln: Kiln
config: Config

dev = FastAPI()
dev.add_middleware(CORSMiddleware,
                   allow_origins=["*"],
                   allow_methods=["*"],
                   allow_headers=["*"])


@dev.on_event("startup")
def load_config():
    global config
    config = tilekiln.load_config(os.environ[TILEKILN_CONFIG])
    config.id = os.environ[TILEKILN_ID]
    # Because the DB connection variables are passed as standard PG* vars,
    # a plain connect() will connect to the right DB

    pool = psycopg_pool.ConnectionPool(min_size=1, max_size=1, num_workers=1,
                                       check=psycopg_pool.ConnectionPool.check_connection)

    global kiln
    kiln = Kiln(config, pool)


@dev.head("/")
@dev.get("/")
def root():
    raise HTTPException(status_code=404)


@dev.head("/favicon.ico")
@dev.get("/favicon.ico")
def favicon():
    return Response("")


@dev.head("/tilejson.json")
@dev.get("/tilejson.json")
def redirect_tilejson():
    global config
    return RedirectResponse(f"/{config.id}/tilejson.json")


@dev.head("/{prefix}/tilejson.json")
@dev.get("/{prefix}/tilejson.json")
def tilejson(prefix):
    global config
    if prefix != config.id:
        raise HTTPException(status_code=404, detail=f"Tileset {prefix} not found on server.")
    return Response(content=config.tilejson(os.environ[TILEKILN_URL]),
                    media_type="application/json",
                    headers=STANDARD_HEADERS)


@dev.head("/{prefix}/{zoom}/{x}/{y}.mvt")
@dev.get("/{prefix}/{zoom}/{x}/{y}.mvt")
def serve_tile(prefix: str, zoom: int, x: int, y:  int):
    global config
    if prefix != config.id:
        raise HTTPException(status_code=404, detail=f"Tileset {prefix} not found on server.")
    global kiln
    return Response(kiln.render(Tile(zoom, x, y)),
                    media_type="application/vnd.mapbox-vector-tile",
                    headers=STANDARD_HEADERS)



'''
Custom exceptions used by tilekiln

Exception (base)
|_ Error
   |_ConfigError
   |_RuntimeError
'''


class Error(Exception):
    pass


class ConfigError(Error):
    pass


class ConfigYAMLError(ConfigError):
    ''' Errors where YAML is invalid, missing, or types are wrong'''
    pass


class ConfigLayerError(ConfigError):
    pass


class DefinitionError(ConfigLayerError):
    pass


class RuntimeError(Error):
    pass


class ZoomNotDefined(RuntimeError):
    pass



import click
import psycopg_pool

import tilekiln
import tilekiln.dev
import tilekiln.server
import tilekiln.scripts.config
import tilekiln.scripts.generate
import tilekiln.scripts.serve
import tilekiln.scripts.storage
from tilekiln.storage import Storage


# Allocated as per https://github.com/prometheus/prometheus/wiki/Default-port-allocations
PROMETHEUS_PORT = 10013


# We want click to print out commands in the order they are defined.
class OrderCommands(click.Group):
    def list_commands(self, ctx: click.Context) -> list[str]:
        return list(self.commands)


@click.group(cls=OrderCommands)
def cli() -> None:
    pass


cli.add_command(tilekiln.scripts.config.config)
cli.add_command(tilekiln.scripts.generate.generate)
cli.add_command(tilekiln.scripts.storage.storage)
cli.add_command(tilekiln.scripts.serve.serve)


@cli.command()
@click.option('--bind-host', default='0.0.0.0', show_default=True,
              help='Bind socket to this host. ')
@click.option('--bind-port', default=PROMETHEUS_PORT, show_default=True,
              type=click.INT, help='Bind socket to this port.')
@click.option('--storage-dbname')
@click.option('--storage-host')
@click.option('--storage-port', type=click.INT)
@click.option('--storage-username')
def prometheus(bind_host: str, bind_port: int, storage_dbname: str, storage_host: str,
               storage_port: int, storage_username: str) -> None:
    '''Run a prometheus exporter for metrics on tiles.'''
    with psycopg_pool.ConnectionPool(min_size=1, max_size=1, num_workers=1,
                                     check=psycopg_pool.ConnectionPool.check_connection,
                                     kwargs={"dbname": storage_dbname,
                                             "host": storage_host,
                                             "port": storage_port,
                                             "user": storage_username}) as pool:
        storage = Storage(pool)

        # tilekiln.prometheus brings in a bunch of stuff, so only do this
        # for this command
        from tilekiln.prometheus import serve_prometheus
        # TODO: make sleep a parameter
        click.echo(f'Running prometheus exporter on http://{bind_host}:{bind_port}/')
        serve_prometheus(storage, bind_host, bind_port, 15)



import psycopg
import psycopg_pool

import tilekiln.errors
from tilekiln.config import Config
from tilekiln.tile import Tile


class Kiln:
    '''
    The kiln is what actually generates the tiles, using the config to compute SQL,
    and a DB connection to execute it
    '''
    def __init__(self, config: Config, pool: psycopg_pool.ConnectionPool):
        self.__config = config
        self.__pool = pool

    def render(self, tile: Tile) -> bytes:
        if tile.zoom < self.__config.minzoom or tile.zoom > self.__config.maxzoom:
            raise tilekiln.errors.ZoomNotDefined

        with self.__pool.connection() as conn:
            with conn.cursor() as curs:
                result = b''
                for sql in self.__config.layer_queries(tile):
                    result += self.__render_layer(curs, sql)

        return result

    def __render_layer(self, curs: psycopg.Cursor, sql: str) -> bytes:
        curs.execute(sql, binary=True)
        for record in curs:
            return record[0]
        raise RuntimeError("No rows in tile query result, should never reach here")



import datetime
import json
import sys
from collections.abc import Iterator

import click
import psycopg.rows
import psycopg_pool

import tilekiln.errors

from tilekiln.metric import Metric
from tilekiln.tile import Tile
from tilekiln.tileset import Tileset

METADATA_TABLE = "metadata"
GENERATE_STATS_TABLE = "generate_stats"
TILE_STATS_TABLE = "tile_stats"

# Lower percentiles are typically not interesting, because generally the
# smallest 50% of tiles are identical water tiles or something similarly
# sparse. Where the data gets interesting is p95 and above.
PERCENTILES = [0.0, 0.25, 0.50, 0.75, 0.90, 0.95, 0.99, 0.999, 1.0]


class Storage:
    '''
    Storage is an object representing a tile storage, backed by a PostgreSQL database

    A Storage contains tiles and metadata about tilesets, and has functions to update
    tiles and metadata based on the ID
    '''
    def __init__(self, pool: psycopg_pool.ConnectionPool, schema="tilekiln"):
        self.__pool = pool
        self.__schema = schema

    '''
    Methods that manipulate schema-related stuff and don't involve any tiles
    '''
    def create_schema(self) -> None:
        with self.__pool.connection() as conn:
            with conn.cursor() as cur:
                # Perform one-time setup using CREATE ... IF NOT EXISTS
                # This is safe to rerun multiple times
                cur.execute(f'''CREATE SCHEMA IF NOT EXISTS "{self.__schema}"''')
                self.__setup_stats(cur)
                self.__setup_metadata(cur)
            conn.commit()

    '''
    Methods for tilesets
    '''
    def create_tileset(self, id: str, minzoom: int, maxzoom: int, tilejson: str) -> None:
        with self.__pool.connection() as conn:
            with conn.cursor() as cur:
                self.__set_metadata(cur, id, minzoom, maxzoom, tilejson)

                self.__setup_tables(cur, id, minzoom, maxzoom)
            conn.commit()

    def remove_tileset(self, id: str) -> None:
        with self.__pool.connection() as conn:
            with conn.cursor() as cur:
                cur.execute(f'''DELETE FROM "{self.__schema}"."{METADATA_TABLE}" WHERE id = %s''',
                            (id,))
                cur.execute(f'''DROP TABLE "{self.__schema}"."{id}" CASCADE''')
                cur.execute(f'''DELETE FROM "{self.__schema}"."{TILE_STATS_TABLE}" WHERE id = %s''',
                            (id,))
            conn.commit()

    def get_tilesets(self) -> Iterator[Tileset]:
        '''
        Gets all tilesets in the storage
        '''

        with self.__pool.connection() as conn:
            with conn.cursor(row_factory=psycopg.rows.dict_row) as cur:
                cur.execute(f'''SELECT id, minzoom, maxzoom, tilejson
                                FROM "{self.__schema}"."{METADATA_TABLE}"''')
                for record in cur:
                    yield Tileset(self, record["id"], record["minzoom"], record["maxzoom"],
                                  json.dumps(record["tilejson"]))

    def get_tileset_ids(self) -> Iterator[str]:
        '''
        Get only the tileset IDs
        '''

        with self.__pool.connection() as conn:
            with conn.cursor(row_factory=psycopg.rows.dict_row) as cur:
                cur.execute(f'''SELECT id
                                FROM "{self.__schema}"."{METADATA_TABLE}"''')
                for record in cur:
                    yield record["id"]

    ''' Methods for metrics'''
    def metrics(self) -> Iterator[Metric]:
        with self.__pool.connection() as conn:
            with conn.cursor(row_factory=psycopg.rows.class_row(Metric)) as cur:
                cur.execute(f'''SELECT id, zoom, num_tiles, size, percentiles
                                FROM "{self.__schema}"."{TILE_STATS_TABLE}"''')
                for record in cur:
                    yield record

    def update_metrics(self) -> None:
        with self.__pool.connection() as conn:
            with conn.cursor() as cur:
                for id in self.get_tileset_ids():
                    minzoom = self.get_minzoom(id)
                    maxzoom = self.get_maxzoom(id)
                    self.__update_tileset_metrics(cur, id, minzoom, maxzoom)
                conn.commit()

    '''Methods that set/get metadata'''
    def set_metadata(self, id, minzoom, maxzoom, tilejson):
        '''
        Saves metadata into storage

        This just wraps __set_metadata, which requires a cursor
        '''
        with self.__pool.connection() as conn:
            with conn.cursor() as cur:
                self.__set_metadata(cur, id, minzoom, maxzoom, tilejson)
                self.__conn.commit()

    # TODO: Should the various get_* functions be separate? The query has to fetch from the
    # DB each time, but only tilejson needs URL. Not an urgent issue.
    def get_tilejson(self, id, url) -> str:
        '''Gets the tilejson for a layer from storage.'''
        with self.__pool.connection() as conn:
            with conn.cursor(row_factory=psycopg.rows.dict_row) as cur:
                cur.execute(f'''SELECT tilejson
                                FROM "{self.__schema}"."{METADATA_TABLE}"
                                WHERE id = %s''', (id,))
                result = cur.fetchone()
                if result is None:
                    # TODO: raise exception and handle it at the calling level
                    click.echo(f"Failed to retrieve tilejson for id {id}, "
                               f"does it exist in storage DB?",
                               err=True)
                    sys.exit(1)
                tilejson = result["tilejson"]
                tilejson["tiles"] = [f"{url}" + "/{z}/{x}/{y}.mvt"]
                return json.dumps(tilejson)

    def get_minzoom(self, id):
        '''Gets the minzoom for a layer from storage.'''
        with self.__pool.connection() as conn:
            with conn.cursor(row_factory=psycopg.rows.dict_row) as cur:
                cur.execute(f'''SELECT minzoom
                                FROM "{self.__schema}"."{METADATA_TABLE}"
                                WHERE id = %s''', (id,))
                result = cur.fetchone()
                if result is None:
                    # TODO: raise exception and handle it at the calling level
                    click.echo(f"Failed to retrieve minzoom for id {id}, "
                               f"does it exist in storage DB?",
                               err=True)
                    sys.exit(1)
                return result["minzoom"]

    def get_maxzoom(self, id):
        '''Gets the minzoom for a layer from storage.'''
        with self.__pool.connection() as conn:
            with conn.cursor(row_factory=psycopg.rows.dict_row) as cur:
                cur.execute(f'''SELECT maxzoom
                                FROM "{self.__schema}"."{METADATA_TABLE}"
                                WHERE id = %s''', (id,))
                result = cur.fetchone()
                if result is None:
                    # TODO: raise exception and handle it at the calling level
                    click.echo(f"Failed to retrieve minzoom for id {id}, "
                               "does it exist in storage DB?", err=True)
                    sys.exit(1)
                return result["maxzoom"]

    '''
    Methods that involve saving, fetching, and deleting tiles
    '''
    def delete_tiles(self, id: str, tiles: set[Tile]):
        with self.__pool.connection() as conn:
            with conn.cursor() as cur:
                for tile in tiles:
                    self.__delete_tile(cur, id, tile)
            conn.commit()

    def truncate_tables(self, id: str, zooms=None):
        with self.__pool.connection() as conn:
            with conn.cursor() as cur:
                if zooms is None:
                    zooms = range(self.get_minzoom(id), self.get_maxzoom(id)+1)
                for zoom in zooms:
                    self.__truncate_table(cur, id, zoom)
            conn.commit()

    def get_tile(self, id: str, tile: Tile) -> tuple[bytes | None, datetime.datetime | None]:
        with self.__pool.connection() as conn:
            with conn.cursor(row_factory=psycopg.rows.dict_row) as cur:
                try:
                    cur.execute(f'''SELECT generated, tile FROM "{self.__schema}"."{id}"
                                    WHERE zoom = %s AND x = %s AND y = %s''',
                                (tile.zoom, tile.x, tile.y), binary=True)
                except psycopg.errors.UndefinedTable:
                    raise tilekiln.errors.ZoomNotDefined

                result = cur.fetchone()
                if result is None:
                    return None, None
                return result["tile"], result["generated"]

    # TODO: Needs to return timestamp written to the DB
    def save_tile(self, id: str, tile: Tile,
                  tiledata: bytes, render_time=0) -> datetime.datetime | None:
        with self.__pool.connection() as conn:
            with conn.cursor(row_factory=psycopg.rows.dict_row) as cur:
                # TODO: This statement unconditionally writes the row even if it's unchanged. It
                # shouldn't. Adding WHERE tile != EXCLUDED.tile would help, but then it would
                # return zero rows if the contents are the same. The method here instead results
                # in extra writes but does preserve the datetime.
                tablename = f"{id}_z{tile.zoom}"
                try:
                    cur.execute(f'''INSERT INTO "{self.__schema}"."{tablename}" AS store\n'''
                                '''(zoom, x, y, tile)\n'''
                                '''VALUES (%s, %s, %s, %s)\n'''
                                '''ON CONFLICT (zoom, x, y)\n'''
                                '''DO UPDATE SET tile = EXCLUDED.tile,\n'''
                                '''generated = CASE WHEN store.tile != EXCLUDED.tile\n'''
                                '''    THEN statement_timestamp()\n'''
                                '''    ELSE store.generated END\n'''
                                '''RETURNING generated''',
                                (tile.zoom, tile.x, tile.y, tiledata))
                except psycopg.errors.UndefinedTable:
                    raise tilekiln.errors.ZoomNotDefined
                result = cur.fetchone()
                if result is None:
                    return None
                return result["generated"]

    def __setup_metadata(self, cur):
        ''' Create the metadata table in storage. This is safe to rerun
        '''
        # TODO: Updating metadata table schema??
        # Probably can only be done on a major version upgrade

        cur.execute(f'''CREATE TABLE IF NOT EXISTS "{self.__schema}"."{METADATA_TABLE}" (\n'''
                    '''id text PRIMARY KEY,\n'''
                    '''active boolean NOT NULL DEFAULT TRUE,\n'''
                    '''minzoom smallint NOT NULL,\n'''
                    '''maxzoom smallint NOT NULL,\n'''
                    '''tilejson jsonb NOT NULL)''')

    def __set_metadata(self, cur, id, minzoom, maxzoom, tilejson):
        '''
        Sets metadata using a cursor

        This is separate from set_metadata because sometimes it needs
        calling within a transaction
        '''
        cur.execute(f'''INSERT INTO "{self.__schema}"."{METADATA_TABLE}"
        (id, minzoom, maxzoom, tilejson)
        VALUES (%s, %s, %s, %s)
        ON CONFLICT (id)
        DO UPDATE SET minzoom = EXCLUDED.minzoom,
        maxzoom = EXCLUDED.maxzoom,
        tilejson = EXCLUDED.tilejson
        ''', (id, minzoom, maxzoom, tilejson))

    def __setup_stats(self, cur):
        '''Create the stats tables.

        One table has tile generation stats, the other has tile storage stats.
        '''
        # Because we're just storing counters for prometheus here and unlogged table is fine.
        # Periodic resets are okay.
        # It's necessary to store this in-db since we might call tilerender more than once
        # in a polling interval.
        # TODO: Use this table
        cur.execute(f'''CREATE UNLOGGED TABLE IF NOT EXISTS
            "{self.__schema}"."{GENERATE_STATS_TABLE}" (
            id text,
            zoom smallint,
            num_rendered integer DEFAULT 0,
            time_rendered interval DEFAULT '0',
            PRIMARY KEY (id, zoom)
        )
        ''')

        # This caches information on the number of tiles. Prometheus can be called every 15 seconds
        # and doing a sequential scan that often is a bad idea
        cur.execute(f'''CREATE TABLE IF NOT EXISTS "{self.__schema}"."{TILE_STATS_TABLE}" (
            id text,
            zoom smallint,
            num_tiles integer NOT NULL,
            size bigint NOT NULL,
            percentiles double precision[][] NOT NULL,
            PRIMARY KEY (id, zoom),
            CHECK (array_length(percentiles, 1) = 2)
        )
        ''')

    def __update_tileset_metrics(self, cur, id, minzoom, maxzoom) -> None:
        for zoom in range(minzoom, maxzoom+1):
            tablename = f"{id}_z{zoom}"
            # This SQL statement needs to handle the case of an empty table.
            # Except for COUNT(*) the aggregate functions return NULL for
            # no rows, which is a problem. One option would be to save
            # {{}, {}} as the array but 2-d empty arrays don't really work
            # in PostgreSQL. Instead, we return 0 for all metrics.
            #
            # We set jit to ON as it is faster when the tables are large, but
            # jit is commonly disabled on tile rendering servers because it
            # slows down rendering queries.
            # TODO: Consider if it would be better to completely skip the row
            #       and emit no metric.
            # TODO: Reformat this statement to be better with line breaks
            cur.execute('SET LOCAL jit TO ON;')
            cur.execute(f'''INSERT INTO "{self.__schema}"."{TILE_STATS_TABLE}"
                        SELECT
                            %(id)s AS id,
                            %(zoom)s AS zoom,
                            COUNT(*) AS num_tiles,
                            COALESCE(SUM(length(tile)),0) AS size,
                            ARRAY[%(percentile)s,
                                COALESCE(PERCENTILE_CONT(%(percentile)s::double precision[])
                                    WITHIN GROUP (ORDER BY length(tile)),
                                    array_fill(0,
                                    ARRAY[array_length(%(percentile)s, 1)]))] AS percentiles
                            FROM "{self.__schema}"."{tablename}"
                            ON CONFLICT (id, zoom)
                        DO UPDATE SET num_tiles = EXCLUDED.num_tiles,
                            size = EXCLUDED.size,
                            percentiles = EXCLUDED.percentiles;
                        ''', {'id': id, 'zoom': zoom, 'percentile': PERCENTILES})

    def __setup_tables(self, cur, id, minzoom, maxzoom):
        '''Create the tile storage tables

        This creates the tile storage tables. It intentionally
        does not try to overwrite existing tables.
        '''
        # These columns are ordered to minimize wasted space between columns
        cur.execute(f'''CREATE TABLE "{self.__schema}"."{id}" (
                    zoom smallint CHECK (zoom >= {minzoom} AND zoom <= {maxzoom}),
                    x int CHECK (x >= 0 AND x < 1 << zoom),
                    y int CHECK (y >= 0 AND y < 1 << zoom),
                    generated timestamptz DEFAULT statement_timestamp(),
                    tile bytea NOT NULL,
                    primary key (zoom, x, y)
                    ) PARTITION BY LIST (zoom)''')
        for zoom in range(minzoom, maxzoom+1):
            tablename = f"{id}_z{zoom}"
            cur.execute(f'''CREATE TABLE "{self.__schema}"."{tablename}"
                            PARTITION OF "{self.__schema}"."{id}"
                            FOR VALUES IN ({zoom})''')

    def __load_metadata(self):
        '''Load the stored metadata.

        This allows serving a TileJSON without having access to the config
        '''
        with self.__conn.cursor(row_factory=psycopg.rows.dict_row) as cur:
            cur.execute(f'''SELECT minzoom, maxzoom, tilejson
                            FROM "{self.__schema}"."{METADATA_TABLE}"
                            WHERE id = %s''', (self.id,))
            result = cur.fetchone()
            if result is None:
                # TODO: raise exception and handle it at the calling level
                click.echo(f"Failed to retrieve metadata for id {self.id}, "
                           "does it exist in storage DB?", err=True)
                sys.exit(1)
            self.minzoom = result["minzoom"]
            self.maxzoom = result["maxzoom"]
            self.__rawtilejson = result["tilejson"]

    def __truncate_table(self, cur, id: str, zoom: int) -> None:
        '''Remove every tile from a particular tileset and zoom'''
        tablename = f"{id}_z{zoom}"
        cur.execute(f'''TRUNCATE TABLE "{self.__schema}"."{tablename}"''')

    def __delete_tile(self, cur, id: str, tile: Tile):
        '''Delete an individual tile

        How this is implemented is not ideal for long lists of tiles
        to delete, but generally a long list is an entire zoom or a box.

        In the former case it is implemented as __truncate_table, and
        the latter case is not implemented but would take min/max x/y.
        '''
        cur.execute(f'''DELETE FROM "{self.__schema}"."{id}"
                        WHERE zoom = %s AND x = %s AND y = %s''',
                    (tile.zoom, tile.x, tile.y))



import pmtiles.tile  # type: ignore


class Tile:
    __slots__ = ("tileid")

    def __init__(self, zoom: int, x: int, y: int):
        '''Creates a tile object, with x, y, and zoom
        '''
        self.tileid = pmtiles.tile.zxy_to_tileid(zoom, x, y)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.tileid == other.tileid

    def __hash__(self):
        return self.tileid

    @property
    def zxy(self):
        return pmtiles.tile.tileid_to_zxy(self.tileid)

    @property
    def zoom(self):
        return self.zxy[0]

    @property
    def x(self):
        return self.zxy[1]

    @property
    def y(self):
        return self.zxy[2]

    def __repr__(self) -> str:
        return f"Tile({self.zoom},{self.x},{self.y})"

    @classmethod
    def from_string(cls, tile: str):
        fragments = tile.split("/")
        return cls(int(fragments[0]), int(fragments[1]), int(fragments[2]))

    @classmethod
    def from_tileid(cls, tileid: int):
        # TODO: This converts id to xyz to id, there should be a way with less calls
        (zoom, x, y) = pmtiles.tile.tileid_to_zxy(tileid)
        return cls(zoom, x, y)

    def bbox(self, buffer) -> str:
        '''Returns the bounding box for a tile
        '''
        return f'''ST_TileEnvelope({self.zoom}, {self.x}, {self.y}, margin=>{buffer})'''



import click

import tilekiln
from tilekiln.tile import Tile


@click.group()
def config() -> None:
    '''Commands to work with and check config files'''
    pass


@config.command()
@click.option('--config', required=True, type=click.Path(exists=True, dir_okay=False))
def test(config: str):
    '''Tests a config for validity.

    The process will exit with exit code 0 if tilekiln can load the config.

    This is intended for build and CI scripts used by configs.
    '''
    tilekiln.load_config(config)
    return 0


@config.command()
@click.option('--config', required=True, type=click.Path(exists=True, dir_okay=False))
@click.option('--layer', type=click.STRING)
@click.option('--zoom', '-z', type=click.INT, required=True)
@click.option('-x', type=click.INT, required=True)
@click.option('-y', type=click.INT, required=True)
def sql(config: str, layer: str, zoom: int, x: int, y: int):
    '''Print the SQL for a tile or layer.

    Prints the SQL that would be issued to generate a particular tile layer,
    or if no layer is given, the entire tile. This allows manual debugging of
    a tile query.
    '''

    c = tilekiln.load_config(config)

    if layer is None:
        for sql in c.layer_queries(Tile(zoom, x, y)):
            click.echo(sql)
        return 0
    else:
        # Iterate through the layers to find the right one
        for lc in c.layers:
            if lc.id == layer:
                sql = lc.render_sql(Tile(zoom, x, y))
                if sql is None:
                    click.echo((f"Zoom {zoom} not between min zoom {lc.minzoom} "
                                f"and max zoom {lc.maxzoom} for layer {layer}."), err=True)
                    return 1
                click.echo(sql)
                return 0
        click.echo(f"Layer '{layer}' not found in configuration", err=True)
        return 1



'''
The code here pulls creates multiple kilns to generate the tiles in parallel
'''
import multiprocessing as mp
from collections.abc import Collection

import psycopg_pool

from tilekiln.config import Config
from tilekiln.kiln import Kiln
from tilekiln.storage import Storage
from tilekiln.tile import Tile
from tilekiln.tileset import Tileset


kiln: Kiln
tileset: Tileset


def setup(config: Config, source_kwargs, storage_kwargs) -> None:  # type: ignore[no-untyped-def]
    '''
    Sets up the kiln and tileset for the worker function.
    '''
    global kiln, tileset
    source_pool = psycopg_pool.ConnectionPool(min_size=1, max_size=1, num_workers=1,
                                              check=psycopg_pool.ConnectionPool.check_connection,
                                              kwargs=source_kwargs)
    kiln = Kiln(config, source_pool)

    storage_pool = psycopg_pool.ConnectionPool(min_size=1, max_size=1, num_workers=1,
                                               check=psycopg_pool.ConnectionPool.check_connection,
                                               kwargs=storage_kwargs)
    storage = Storage(storage_pool)
    tileset = Tileset.from_config(storage, config)


def worker(tile: Tile) -> None:
    global kiln, tileset
    mvt = kiln.render(tile)
    tileset.save_tile(tile, mvt)


def generate(config: Config, source_kwargs, storage_kwargs,  # type: ignore[no-untyped-def]
             tiles: Collection[Tile], num_processes: int) -> None:

    # If there are no processes and no tiles then there's nothing to do.
    if num_processes == 0 and len(tiles) == 0:
        return

    with mp.Pool(num_processes, setup, (config, source_kwargs, storage_kwargs)) as pool:
        pool.imap_unordered(worker, tiles)
        pool.close()
        pool.join()



from tilekiln.tile import Tile


class Tilerange():
    def __init__(self, minz, maxz):
        self.minid = Tile(minz, 0, 0).tileid
        self.maxid = Tile(maxz + 1, 0, 0).tileid

    def __iter__(self):
        for id in range(self.minid, self.maxid):
            yield Tile.from_tileid(id)

    def __len__(self):
        return self.maxid - self.minid

    def __contains__(self, value):
        raise NotImplementedError



import time

import prometheus_client
from prometheus_client.registry import Collector
from prometheus_client.core import GaugeMetricFamily, REGISTRY

from tilekiln.storage import Storage

# Disable default metrics since we're not monitoring this process, we're monitoring
# the DB sizes
REGISTRY.unregister(prometheus_client.GC_COLLECTOR)
REGISTRY.unregister(prometheus_client.PLATFORM_COLLECTOR)
REGISTRY.unregister(prometheus_client.PROCESS_COLLECTOR)

# Don't auto-create metrics for when the metric was created
prometheus_client.disable_created_metrics()


class TilekilnCollector(Collector):
    def __init__(self, storage: Storage):
        self.__storage = storage
        super().__init__()

        self.__i = 0

    # This one is run every 15s
    def collect(self):
        # This is manually producing the metrics described in
        # https://prometheus.io/docs/concepts/metric_types/#summary
        # Native histograms would be nice here, but are still only experimental
        size = GaugeMetricFamily('tilekiln_stored_bytes_sum', 'Total size of tiles',
                                 labels=['tileset', 'zoom'])
        quantiles = GaugeMetricFamily('tilekiln_stored_bytes', 'Tile percentiles',
                                      labels=['tileset', 'zoom', 'quantile'])
        total = GaugeMetricFamily('tilekiln_stored_count', 'Tiles in tilekiln storage',
                                  labels=['tileset', 'zoom'])
        for metric in self.__storage.metrics():
            size.add_metric([metric.id, str(metric.zoom)], metric.size)
            total.add_metric([metric.id, str(metric.zoom)], metric.num_tiles)
            for i in range(0, len(metric.percentiles[0])):
                quantiles.add_metric([metric.id, str(metric.zoom), str(metric.percentiles[0][i])],
                                     metric.percentiles[1][i])
        yield total
        yield size
        yield quantiles

    def update(self):
        self.__i = self.__i + 1
        pass


METRIC_UPDATE_TIME = prometheus_client.Summary('tilekiln_metrics_storage_seconds',
                                               'Time spent updating metrics')


@METRIC_UPDATE_TIME.time()
def monitored_update_metrics(storage: Storage):
    '''Update storage metrics while tracking call time

    The easiest way to monitor a function is to annotate it. Rather than require
    prometheus in storage.py, we wrap it and annotate the wrapper to track call time.
    '''
    storage.update_metrics()


def serve_prometheus(storage: Storage, addr, port, sleep):
    '''Start a prometheus server for storage info.'''
    collector = TilekilnCollector(storage)
    REGISTRY.register(collector)
    # start http
    prometheus_client.start_http_server(port=port, addr=addr)
    while True:
        # TODO: Time this with prometheus
        monitored_update_metrics(storage)
        time.sleep(sleep)
